import toml


class TomlReader:

    @classmethod
    def of(cls, path: str) -> dict:
        """
        Open a toml file.

        Args:
            path (``str``): Path to the toml file.

        Returns:
            ``dict``: The opened toml file.

        """
        return toml.load(path)

    def __new__(cls, path: str):
        return cls.of(path)



import importlib.util
import os
import time
from functools import reduce
from os.path import dirname, abspath
from types import ModuleType
from typing import Callable, List, Dict, Iterable, Sequence, Any, Optional, Type, Union, Tuple

import numpy as np

from .fp import F, _, Stream
from .type import T, R, T_E


def identity(*args, **kwargs):
    """
    Pickle-friendly identity function.
    """
    assert kwargs == {}, "identity function doesn't accept keyword arguments"
    if len(args) == 1:
        return args[0]
    else:
        return args


def generate_inf_seq(items: Iterable[T]) -> Stream[T]:
    """
    Generate an infinity late-evaluate sequence.

    Args:
        items [``Iterable[T]``]: Original items.

    Returns:
        :class:`~fp.stream.Stream[T]`: The infinity sequence.

    Examples::

        >>> seq = generate_inf_seq(["a", 1, ["e0", "e1"]])
        >>> list(seq[:10])
        ['a', 1, ['e0', 'e1'], 'a', 1, ['e0', 'e1'], 'a', 1, ['e0', 'e1'], 'a']

    Note:
        The repeated reference type objects are the same object.

        >>> a_list = ["a", "b"]
        >>> seq = list(generate_inf_seq([a_list])[:3])
        >>> seq
        [['a', 'b'], ['a', 'b'], ['a', 'b']]
        >>> a_list.append("c")
        >>> seq
        [['a', 'b', 'c'], ['a', 'b', 'c'], ['a', 'b', 'c']]

    """
    s = Stream()
    return s << items << map(_, s)


def compose(fs: Sequence[Callable]) -> F:
    """
    Compose functions as a pipeline function.

    Args:
        fs (``Sequence[Callable]``): The functions input for composition.

    Returns:
        :class:`~fp.func.F`: The composed output function.

    Examples::

        >>> f = lambda x: x + 1
        >>> g = lambda x: x * 2
        >>> h = lambda x: x ** 2
        >>> x = 1
        >>> h(g(f(x))) == compose([f, g, h])(x)
        True

    """
    return reduce(_ >> _, fs, F())


def listdir(path: str, filter_func: Callable[[str], bool] = lambda arg: True) -> List[str]:
    """
    The full path version of :func:`os.listdir`.

    Get the all listdir result with input path.

    Args:
        path (``str``): The path for listdir.
        filter_func (``(str) -> bool``, optional): The filter function to filter the file/directory name in ``listdir``.

    Returns:
        ``List[str]``: listdir result with input path.

    Examples::

        >>> os.listdir("tensorneko/util")[:3]  # python os library
        ['configuration.py', 'enum.py', 'string_getter.py']
        >>> listdir("tensorneko/util")[:3]  # tensorneko
        ['tensorneko/util/configuration.py',
         'tensorneko/util/enum.py',
         'tensorneko/util/string_getter.py']

    """
    files = filter(filter_func, os.listdir(path))
    return list(map(F(os.path.join, path), files))


def with_printed(x: T, func: Callable[[T], Any] = identity) -> T:
    """
    An identity function but with printed to console with some transform.

    Args:
        x (``T``): Input.
        func (``(T) -> Any``, optional): A function used to apply the input for printing. Default: identity function.

    Returns:
        ``T``: Identity output.

    Examples::

        >>> x = torch.tensor([1.5, 2.5, 3.5])
        >>> y = with_printed(x, lambda tensor: tensor.dtype)
        torch.float32
        >>> x == y
        tensor([True, True, True])

    """
    print(func(x))
    return x


def ifelse(predicate: Callable[[T], bool], func_true: Callable[[T], R], func_false: Callable[[T], R] = identity
) -> Callable[[T], R]:
    """
    A function composition util for if-else control flow.

    Args:
        predicate (``(T) -> bool``): A predicate produce a bool.
        func_true (``(T) -> R``): If the bool from predicate is True, return this function.
        func_false (``(T) -> R``): If the bool from predicate is False, return this function.

    Returns:
        ``(T) -> R``: Composed function with if-else control flow.
    """

    def wrapper(*args, **kwargs):
        if predicate(*args, **kwargs):
            return func_true(*args, **kwargs)
        else:
            return func_false(*args, **kwargs)

    return wrapper


def dict_add(*dicts: dict) -> dict:
    """
    Merge multiple dictionaries.

    Args:
        *dicts: dictionaries to merge.

    Returns:
        ``dict``: The merged dictionary.
    """
    new_dict = {}
    for each in dicts:
        new_dict.update(each)
    return new_dict


def as_list(*args, **kwargs) -> list:
    """
    Returns:
        ``list``: The list of args.

    Examples::

        >>> as_list(1, 2, 3, 4, key=5)
        [1, 2, 3, 4, 5]
    """
    return list(args) + list(kwargs.values())


def list_to_dict(l: List[T], key: Callable[[T], R]) -> Dict[R, T]:
    """
    Convert the list as a dictionary by given key.
    Args:
        l (``List[T]``): Input list.
        key (``(T) -> R``): The key getter function.

    Returns:
        ``Dict[R, T]``: The output dictionary.
    """
    return {key(x): x for x in l}


def get_tensorneko_util_path() -> str:
    """
    Get the `tensorneko_util` library root path

    Returns:
        ``str``: The root path of `tensorneko`
    """
    return dirname(dirname(abspath(__file__)))


def circular_pad(x: List[T], target: int) -> List[T]:
    """
    Circular padding a list to the target length.

    Args:
        x(``List[T]``): Input list.
        target(``int``): Target length.

    Returns:
        ``List[T]``: The padded list.
    """
    if len(x) == target:
        return x
    elif len(x) > target:
        return x[:target]
    elif 2 * len(x) > target:
        return x + x[:target - len(x)]
    else:
        return circular_pad(x + x, target)


def load_py(path: str) -> ModuleType:
    """
    Load a python file as a module.

    Args:
        path (``str``): The path of the python file.

    Returns:
        ``Any``: The loaded module.
    """
    spec = importlib.util.spec_from_file_location("module.name", path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def try_until_success(func: Callable[[Any], T], *args, max_trials: Optional[int] = None, sleep_time: int = 0,
    exception_type: Union[Type[T_E], Tuple[T_E, ...]] = Exception, **kwargs
) -> T:
    """
    Try to run the function until success.

    Args:
        func (``(...) -> T``): The function to run.
        max_trials (``int``, optional): The max try times. None for unlimited. Default: None.
        sleep_time (``int``, optional): The sleep time between each try. Default: 0.
        exception_type (``Type[Exception] | (Type[Exception], ...)``, optional): The exception types to catch.
            Default: Exception.
        *args: The args for the function.
        **kwargs: The kwargs for the function.

    Returns:
        ``T``: The output of the function.

    Raises:
        ``Exception``: The exception from the function.
    """
    trials = 0
    while True:
        try:
            return func(*args, **kwargs)
        except exception_type as e:
            trials += 1
            if max_trials is not None and trials > max_trials:
                raise e
        if sleep_time > 0:
            time.sleep(sleep_time)


def sample_indexes(total_frames: int, n_frames: int, sample_rate: int) -> np.ndarray:
    """
    Sample continuous indexes from total frames.

    Args:
        total_frames (``int``): The total number of items from source.
        n_frames (``int``): The number of items to sample.
        sample_rate (``int``): The sample rate to retrieve the continuous items.
    Returns:
        ``np.ndarray``: The sampled indexes.

    Examples::

        >>> sample_indexes(10, 3, 1)
        array([5, 6, 7])

        >>> sample_indexes(10, 3, 2)
        array([1, 3, 5])

        >>> sample_indexes(10, 3, 3)
        array([2, 5, 8])
4
    """

    start_ind = np.random.randint(0, total_frames - (n_frames * sample_rate) + 2, ())
    return np.arange(n_frames) * sample_rate + start_ind



from __future__ import annotations

from logging import Logger


class DummyLogger(Logger):
    """
    A dummy logger that prints all messages to stdout. Or choose to not print anything.
    It is useful when you want to disable logging in a library.

    Args:
        verbose (``bool``): Whether to print the message to stdout.

    Examples::

        >>> import logging
        >>> from tensorneko_util.debug import DummyLogger
        >>> logger = logging.getLogger("annoy_logger", verbose=False)
        >>> DummyLogger.convert(logger)
        >>> logger.info("hello info")
        >>> logger.debug("hello debug")
        >>> logger.warning("hello warning")
        >>> # no output

    """

    def __init__(self, verbose: bool):
        self.verbose = verbose

    @classmethod
    def convert(cls, logger: Logger, verbose: bool = False) -> None:
        """Build a dummy logger from a real logger. It's an inplace operation."""
        cls.__init__(logger, verbose)
        logger.__class__ = cls

    def debug(
        self,
        msg,
        *args,
        exc_info=...,
        stack_info=...,
        stacklevel=...,
        extra=...,
    ):
        if self.verbose:
            print(msg)

    def info(
        self,
        msg,
        *args,
        exc_info=...,
        stack_info=...,
        stacklevel=...,
        extra=...,
    ):
        if self.verbose:
            print(msg)

    def warning(
        self,
        msg,
        *args,
        exc_info=...,
        stack_info=...,
        stacklevel=...,
        extra=...,
    ):
        if self.verbose:
            print(msg)

    def error(
        self,
        msg,
        *args,
        exc_info=...,
        stack_info=...,
        stacklevel=...,
        extra=...,
    ):
        if self.verbose:
            print(msg)

    def critical(
        self,
        msg,
        *args,
        exc_info=...,
        stack_info=...,
        stacklevel=...,
        extra=...,
    ):
        if self.verbose:
            print(msg)



from __future__ import annotations

import time
from functools import wraps
from typing import Callable, Optional

from .type import T


class Timer:
    """
    A timer to count the time of a function or a block of code.

    Examples::

        from tensorneko.util import Timer
        import time

        # use as a context manager with single time
        with Timer():
            time.sleep(1)

        # use as a context manager with multiple segments
        with Timer() as t:
            time.sleep(1)
            t.time("sleep A")
            time.sleep(1)
            t.time("sleep B")
            time.sleep(1)

        # use as a decorator
        @Timer()
        def f():
            time.sleep(1)
            print("f")

        # without parentheses
        @Timer
        def g():
            time.sleep(1)
            print("f")

        # disable verbose and get time manually
        with Timer(verbose=False) as t:
            time.sleep(1)
            dt = t.time("sleep A")
            print(f"Time of sleep A: {dt}")
            time.sleep(1)
            dt = t.time("sleep B")
            print(f"Time of sleep B: {dt}")
            print(f"Time of sleep A and B: {t.elapsed}")
            time.sleep(1)
            print(f"Total time: {t.elapsed}")

    """

    def __init__(self, verbose: bool = True):
        self.times = []
        self.names = []
        self.total_time = None
        self.verbose = verbose

    def __new__(cls, *args, **kwargs):
        if len(args) > 0 and isinstance(args[0], Callable):
            return cls()(args[0])
        else:
            return super().__new__(cls)

    def __enter__(self):
        self.times.append(time.perf_counter())
        return self

    @property
    def elapsed(self):
        return time.perf_counter() - self.times[0]

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.total_time = self.elapsed
        if self.verbose:
            print(self._make_str("Total", self.total_time))

    def time(self, name: Optional[str] = None) -> float:
        self.times.append(time.perf_counter())
        name = name or ""
        dt = time.perf_counter() - self.times[-2]
        if self.verbose:
            print(self._make_str(name, dt))
        return dt

    def __call__(self, f: Callable):
        @wraps(f)
        def wrapper(*args, **kwargs):
            self.times.append(time.perf_counter())
            result = f(*args, **kwargs)
            print(self._make_str(f.__name__, time.perf_counter() - self.times[-1]))
            return result

        return wrapper

    @staticmethod
    def _make_str(name: str, t: float):
        return f"[Timer] {name}: {t} sec"



from typing import Type

from .audio import AudioWriter
from .image import ImageWriter
from .json import JsonWriter
from .pickle import PickleWriter
from .text import TextWriter
from .video import VideoWriter

try:
    from .matlab import MatWriter
    scipy_available = True
except ImportError:
    scipy_available = False
    MatWriter = None

try:
    from .yaml import YamlWriter
    yaml_available = True
except ImportError:
    YamlWriter = object
    yaml_available = False

try:
    from .hdf5 import Hdf5Writer
    h5py_available = True
except ImportError:
    Hdf5Writer = object
    h5py_available = False

try:
    from .toml import TomlWriter
    toml_available = True
except ImportError:
    TomlWriter = object
    toml_available = False


class Writer:

    def __init__(self):
        self.image = ImageWriter
        self.text = TextWriter
        self.json = JsonWriter
        self.video = VideoWriter
        self.audio = AudioWriter
        self.pickle = PickleWriter
        self._mat = None
        self._yaml = None
        self._h5 = None
        self._toml = None

    @property
    def mat(self) -> Type[MatWriter]:
        if scipy_available:
            if self._mat is None:
                self._mat = MatWriter
            return self._mat
        else:
            raise ImportError("Scipy is not available to write mat file.")

    @property
    def yaml(self) -> Type[YamlWriter]:
        if yaml_available:
            if self._yaml is None:
                self._yaml = YamlWriter
            return self._yaml
        else:
            raise ImportError("To use the yaml writer, please install pyyaml.")

    @property
    def h5(self) -> Type[Hdf5Writer]:
        if h5py_available:
            if self._h5 is None:
                self._h5 = Hdf5Writer
            return self._h5
        else:
            raise ImportError("To use the hdf5 writer, please install h5py.")

    @property
    def toml(self) -> Type[TomlWriter]:
        if toml_available:
            if self._toml is None:
                self._toml = TomlWriter
            return self._toml
        else:
            raise ImportError("To use the toml writer, please install toml.")

    def __call__(self, path: str, obj, *args, **kwargs):
        """Automatically infer the file type and return the corresponding result. """
        if path.endswith(".jpg") or path.endswith(".jpeg") or path.endswith(".png") or path.endswith(".bmp"):
            self.image(path, obj, *args, **kwargs)
        elif path.endswith(".txt"):
            return self.text(path, obj, *args, **kwargs)
        elif path.endswith(".json"):
            return self.json(path, obj, *args, **kwargs)
        elif path.endswith(".mp4") or path.endswith(".avi") or path.endswith(".mov") or path.endswith(".mkv"):
            return self.video(path, obj, *args, **kwargs)
        elif path.endswith(".wav") or path.endswith(".mp3") or path.endswith(".flac"):
            return self.audio(path, obj, *args, **kwargs)
        elif path.endswith(".pkl") or path.endswith(".pickle"):
            assert len(args) == 0 and len(kwargs) == 0, "Pickle reader does not support extra arguments."
            return self.pickle(path, obj)
        elif path.endswith(".mat"):
            assert len(args) == 0 and len(kwargs) == 0, "Mat reader does not support extra arguments."
            return self.mat(path, obj)
        elif path.endswith(".yaml") or path.endswith(".yml"):
            assert len(args) == 0 and len(kwargs) == 0, "Yaml reader does not support extra arguments."
            return self.yaml(path, obj)
        elif path.endswith(".h5") or path.endswith(".hdf5"):
            raise NotImplementedError("Hdf5 writer is not implemented yet.")
        elif path.endswith(".toml"):
            assert len(args) == 0 and len(kwargs) == 0, "Toml reader does not support extra arguments."
            return self.toml(path, obj)
        else:
            raise ValueError("Unknown file type: {}".format(path))



from __future__ import annotations

import http.server
from abc import ABC, abstractmethod
from http.server import HTTPServer
from threading import Thread
from typing import Optional, List


class AbstractServer(ABC):

    @abstractmethod
    def start(self):
        ...

    @abstractmethod
    def stop(self):
        ...

    def __enter__(self) -> AbstractServer:
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


class HttpThreadServer(AbstractServer, ABC):
    servers: List[HttpThreadServer] = []

    def __init__(self, path: str, port: int = 8000, name: str = "tensorneko server"):
        self.server_dir = path
        self.port = port
        self.name = name
        self.process: Optional[Thread] = None
        self.httpd: Optional[HTTPServer] = None

        class HTTPHandler(http.server.SimpleHTTPRequestHandler):

            def __init__(self, *args, **kwargs):
                super().__init__(*args, directory=path, **kwargs)

            def log_message(self, format: str, *args) -> None:
                return

        self.Handler = HTTPHandler
        self.Handler.extensions_map[".js"] = "application/javascript"
        self.Handler.extensions_map[".mjs"] = "application/javascript"
        HTTPServer.allow_reuse_address = True
        HTTPServer.allow_reuse_port = True
        self.servers.append(self)

    def start(self) -> None:
        """
        Start or restart the server.
        """
        # stop the old server
        if self.process is not None:
            self.stop()
        # copy files to the view directory
        self._prepare()
        # run python http server
        self._run()

    def start_blocking(self) -> None:
        """
        Start or restart the server on blocking.
        """
        # stop the old server
        if self.process is not None:
            self.stop()
        # copy files to the view directory
        self._prepare()
        # run python http server
        self._run_blocking()

    @abstractmethod
    def _prepare(self) -> None:
        pass

    def _run(self) -> None:
        self.httpd = http.server.HTTPServer(("", self.port), self.Handler)
        self.process = Thread(target=self.httpd.serve_forever)
        self.process.start()
        print(self.server_start_string)

    def _run_blocking(self):
        with http.server.HTTPServer(("", self.port), self.Handler) as httpd:
            print(self.server_start_string)
            httpd.serve_forever()

    def stop(self) -> None:
        """
        Stop the server.
        """
        if self.process is not None:
            self.httpd.shutdown()
            self.process.join()
            print("Server stopped")
            self.process = None
        else:
            print("Server is stopped.")

    @classmethod
    def stop_all(cls) -> None:
        """
        Stop all servers.
        """
        for server in cls.servers:
            server.stop()

    @property
    def server_start_string(self) -> str:
        return f"[Server] {self.name} started at http://127.0.0.1:{self.port}/."



from enum import Enum
from typing import Optional

from .visual_lib import VisualLib


class AudioLib(Enum):
    PYTORCH = 1
    FFMPEG = 2

    _is_torchaudio_available: Optional[bool] = None

    @classmethod
    def pytorch_available(cls) -> bool:
        if cls._is_torchaudio_available is None:
            try:
                import torchaudio
                cls._is_torchaudio_available = True
            except ImportError:
                cls._is_torchaudio_available = False
        return cls._is_torchaudio_available

    @classmethod
    def ffmpeg_available(cls) -> bool:
        return VisualLib.ffmpeg_available()
